/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.feature.auth.jaas;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.AppConfigurationEntry;
import javax.security.auth.login.Configuration;
import javax.security.auth.login.AppConfigurationEntry.LoginModuleControlFlag;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;

import org.bluestemsoftware.open.eoa.commons.util.DOMUtils;
import org.bluestemsoftware.open.eoa.commons.util.DOMValidator;
import org.bluestemsoftware.open.eoa.ext.feature.auth.jaas.util.Constants;
import org.bluestemsoftware.specification.eoa.DeploymentContext;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.Resource;
import org.bluestemsoftware.specification.eoa.ext.Extension;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.jaas.LoginContextFeature;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.jaas.LoginContextFeatureException;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

public final class LoginContextFeatureImpl implements LoginContextFeature.Provider {

    private static final long serialVersionUID = 1L;

    private static final Log log = SystemContext.getContext().getSystem().getLog(LoginContextFeature.class);

    public static final String IMPL = LoginContextFeatureImpl.class.getName();

    private DeploymentContext deploymentContext;
    private Element configuration;

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.Feature$Provider#spi_init(org.w3c.dom.Element)
     */
    public void spi_init() throws LoginContextFeatureException {

        log.debug("init begin");

        log.info("setting up jass configuration");

        if (configuration == null) {
            String loc = "classpath:///schema/http.bluestemsoftware.org.open.eoa.ext.feature.auth.jaas.config.1.0.xml";
            try {
                Resource resource = deploymentContext.getResource(loc);
                InputSource inputSource = new InputSource(resource.getInputStream());
                inputSource.setSystemId(loc);
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                factory.setNamespaceAware(true);
                DocumentBuilder documentBuilder = factory.newDocumentBuilder();
                documentBuilder.setEntityResolver(deploymentContext);
                configuration = documentBuilder.parse(inputSource).getDocumentElement();
            } catch (Exception ex) {
                throw new LoginContextFeatureException("Error retrieving default configuration. " + ex);
            }
        }

        validateConfiguration(new DOMSource(configuration));

        ConfigurationImpl jaasConfiguration = new ConfigurationImpl();

        QName childName = new QName(Constants.FEATURE_SCHEMA_NS, "apps");
        Element appsElement = DOMUtils.getChildElement(configuration, childName);
        childName = new QName(Constants.FEATURE_SCHEMA_NS, "app");
        List<Element> appElements = DOMUtils.getChildElements(appsElement, childName);
        for (Element appElement : appElements) {
            String appName = appElement.getAttribute("name");
            childName = new QName(Constants.FEATURE_SCHEMA_NS, "entries");
            Element entriesElement = DOMUtils.getChildElement(appElement, childName);
            childName = new QName(Constants.FEATURE_SCHEMA_NS, "entry");
            List<Element> entryElements = DOMUtils.getChildElements(entriesElement, childName);
            List<AppConfigurationEntry> entry = new ArrayList<AppConfigurationEntry>();
            for (Element entryElement : entryElements) {
                childName = new QName(Constants.FEATURE_SCHEMA_NS, "module");
                Element moduleElement = DOMUtils.getChildElement(entryElement, childName);
                String className = DOMUtils.getText(moduleElement);
                childName = new QName(Constants.FEATURE_SCHEMA_NS, "flag");
                Element flagElement = DOMUtils.getChildElement(entryElement, childName);
                LoginModuleControlFlag flag = valueOf(DOMUtils.getText(flagElement));
                childName = new QName(Constants.FEATURE_SCHEMA_NS, "options");
                Element optionsElement = DOMUtils.getChildElement(entryElement, childName);
                childName = new QName(Constants.FEATURE_SCHEMA_NS, "option");
                List<Element> optionElements = DOMUtils.getChildElements(optionsElement, childName);
                Map<String, String> options = new HashMap<String, String>();
                for (Element optionElement : optionElements) {
                    String name = optionElement.getAttribute("name");
                    String value = optionElement.getAttribute("value");
                    options.put(name, value);
                }
                entry.add(new AppConfigurationEntry(className, flag, options));
            }
            jaasConfiguration.addEntry(appName, entry);
        }

        Configuration.setConfiguration(jaasConfiguration);

        log.info("jaas configuration complete");

        log.debug("init end");

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.Feature$Provider#spi_destroy()
     */
    public void spi_destroy() {
        log.debug("destroy begin");
        log.debug("destroy end");
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.Extension$Provider#spi_setConsumer(org.bluestemsoftware.specification.eoa.ext.Extension.Consumer)
     */
    public void spi_setConsumer(Extension consumer) {
        this.deploymentContext = consumer.getExtensionFactory().getFactoryContext();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.Feature$Provider#spi_setConfiguration(org.w3c.dom.Element)
     */
    public void spi_setConfiguration(Element configuration) {
        this.configuration = configuration;
    }

    private void validateConfiguration(DOMSource domSource) throws LoginContextFeatureException {

        Resource resource = null;
        try {
            resource = deploymentContext.getResource(Constants.FEATURE_SCHEMA_LOC);
        } catch (DeploymentException de) {
            throw new LoginContextFeatureException("Error validating configuration. " + de);
        }

        // note that we instantiate schema factory impl directly. a bug in jdk 1.5
        // prevents discovery using jaxp newInstance method. and ... the default
        // jdk 1.5 impl has a bug re: parsing qnames, i.e. a bogus UndeclaredPrefix
        // error

        javax.xml.validation.SchemaFactory schemaFactory = null;
        schemaFactory = new org.apache.xerces.jaxp.validation.XMLSchemaFactory();
        javax.xml.validation.Schema schema = null;
        try {
            schema = schemaFactory.newSchema(new StreamSource(resource.getInputStream()));
        } catch (Exception ex) {
            throw new LoginContextFeatureException("Error parsing configuration schema. " + ex);
        }

        DOMValidator domValidator = new DOMValidator(schema);
        String validationError;
        try {
            validationError = domValidator.validate(domSource);
        } catch (Exception ex) {
            throw new LoginContextFeatureException("Error validating configuration. " + ex);
        }
        if (validationError != null) {
            throw new LoginContextFeatureException("Configuration invalid. " + validationError);
        }

    }

    private LoginModuleControlFlag valueOf(String flag) {
        if (flag.toLowerCase().equals("required")) {
            return LoginModuleControlFlag.REQUIRED;
        }
        if (flag.toLowerCase().equals("requisite")) {
            return LoginModuleControlFlag.REQUISITE;
        }
        if (flag.toLowerCase().equals("sufficient")) {
            return LoginModuleControlFlag.SUFFICIENT;
        }
        if (flag.toLowerCase().equals("optional")) {
            return LoginModuleControlFlag.OPTIONAL;
        }
        throw new IllegalArgumentException("flag " + flag + " unrecognized.");
    }

    class CallbackHandlerImpl implements CallbackHandler {

        public void handle(Callback[] callbacks) throws IOException, UnsupportedCallbackException {
            for (int i = 0; i < callbacks.length; i++) {
                if (callbacks[i] instanceof NameCallback) {
                    ((NameCallback)callbacks[i]).setName(null);
                } else if (callbacks[i] instanceof PasswordCallback) {
                    ((PasswordCallback)callbacks[i]).setPassword(null);
                } else {
                    throw new UnsupportedCallbackException(callbacks[i]);
                }
            }
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.Feature$Provider#spi_getFeatureImpl()
     */
    public String spi_getFeatureImpl() {
        return IMPL;
    }

}
